/* $Id: game_engine.c,v 1.22 2006/01/28 02:28:34 ppl Exp $ */

/*
	Control the state of the game. Controls if the behavior of
	the worm. Determines if the worm hit a wall or an obstacle.
*/

#include "xsfcave.h"

#include "game_engine.h"
#include "replay.h"

#define OBSTACLE_WIDTH   4
#define OBSTACLE_HEIGHT 15

static unsigned long diff_timeval(const struct timeval *time1, const struct timeval *time0);
static unsigned int  game_random (unsigned int n, unsigned int m);
static void          game_tick   (GameClass *game, Bool go_up, ReplayClass *);
static long          longrand(void);
static void          update_worm (GameClass *);

void
game_create(GameClass *game)
{
	game->tick_count = 0;

	game->worm_y = GAME_WIDTH / 2;

	game->worm_weight = GAME_DEFAULT_WORM_WEIGHT;

	game->worm_haccel = -20*game->worm_weight;

#ifdef DEBUG
	if (wflag == True)
	{
		game->worm_weight = wflag_arg;
	}
#endif

	game->tunel_height = GAME_WIDTH - 4;

	game->game_over = False;

	game->wall_height = 2;

	/*
		Worm
	*/
	game->worm_count = 0;
	update_worm(game);

	/*
		Starting wall is 2 pixel straight line
	*/
	game->wall_top[0].x = GAME_WIDTH + 1;
	game->wall_top[0].y = -1;

	game->wall_top[1].x = -1;
	game->wall_top[1].y = -1;

	game->wall_top[2].x = 0;
	game->wall_top[2].y = 2;

	game->wall_top[3].x = GAME_WIDTH + 5;
	game->wall_top[3].y = 2;

	/*
		Same for bottom
	*/
	game->wall_bottom[0].x = GAME_WIDTH + 1;
	game->wall_bottom[0].y = GAME_HEIGHT;

	game->wall_bottom[1].x = -1;
	game->wall_bottom[1].y = GAME_HEIGHT;

	game->wall_bottom[2].x = 0;
	game->wall_bottom[2].y = 2 + game->tunel_height;

	game->wall_bottom[3].x = GAME_WIDTH + 5;
	game->wall_bottom[3].y = 2 + game->tunel_height;

	game->wall_count = 4;

	/*
		Obstacles
	*/
	game->obstacle_count = 0;

	game->obstacle_milestone = (double) GAME_WIDTH / 100 * 20;
}

void
game_destroy(GameClass *game)
{
	/*
		There's currently no dynamic data to free in a
		`GameClass'.
	*/
}

unsigned int
game_next_tick_in(const GameClass *game)
{
	struct timeval tv;
	unsigned long diff;

	if (game->tick_count == 0)
	{
		return TICK_LENGTH;
	}

	gettimeofday(&tv, NULL);

	diff = diff_timeval(&tv, &game->last_tick);

	if (diff > TICK_LENGTH)
	{
		return 0;
	}

	return TICK_LENGTH - diff;
}

#define SCORE(x) (((x)->tick_count) / 2)

unsigned
game_score(const GameClass *game)
{
	return SCORE(game);
}

unsigned
game_score_replay(const ReplayClass *replay)
{
	return SCORE(replay);
}

#undef SCORE

void
game_update_ticks(GameClass *game, Bool go_up, ReplayClass *replay)
{
	if (game->tick_count == 0)
	{
		gettimeofday(&game->last_tick, NULL);
		game_tick(game, go_up, replay);
	}
	else
	{
		struct timeval tv;

		gettimeofday(&tv, NULL);

		while (1)
		{
			unsigned long diff;

			diff = diff_timeval(&tv, &game->last_tick);

			if (diff > TICK_LENGTH)
			{
				game->last_tick.tv_usec += TICK_LENGTH;

				if (game->last_tick.tv_usec >= 1000000)
				{
					game->last_tick.tv_sec  += 1;
					game->last_tick.tv_usec -= 1000000;
				}

				game_tick(game, go_up, replay);

				/* DEBUGGING */
				if (replay == NULL) return;
			}
			else
			{
				return;
			}
		}
	}
}

/*
	More then one ticks may have passed since this was called.
*/
static void
game_tick(GameClass *game, Bool go_up, ReplayClass *replay)
{
	int i;

	/* I wonder if people will get *that* addicted */
	if (game->tunel_height == 0)
	{
		game->game_over = True;
	}

	/*
		Calling game_tick after death have no effect.
	*/
	if (game->game_over == True)
	{
		return;
	}

	game->tick_count += 1;
	game->obstacle_milestone -= 1;

	if (replay != NULL)
	{
		Replay.add_tick(replay, go_up);
	}

	if (game->tick_count % 50 == 0)
	{
		game->wall_height += 1;
		game->tunel_height -= 1;
	}

	if (go_up) 
	{
		game->worm_haccel -= game->worm_weight;
	}
	else
	{
		game->worm_haccel += game->worm_weight;
	}

	game->worm_y += game->worm_haccel;

	/*
		Move everything one pixel left
	*/

	for (i = 0; i < game->worm_count; ++i)
	{
		game->worm[i].x1 -= 1;
		game->worm[i].x2 -= 1;
	}

	for (i = 2; i < game->wall_count; ++i)
	{
		game->wall_top[i].x -= 1;
		game->wall_bottom[i].x -= 1;
	}

	for (i = 0; i < game->obstacle_count; ++i)
	{
		game->obstacle[i].x -= 1;
	}

	/*
		Worm doesn't need any removal. The array cycle on
		itself for ever.
	*/

	/*
		Remove old walls

		wall[3] is the next point that will replace the
		left-most point

		     0        1          2        3
		[reserved][reserved][left-most][ next ][ ... ]
	*/
	if (game->wall_top[3].x < 0)
	{
		size_t size;

		game->wall_count -= 1;

		size = (game->wall_count - 2) * sizeof game->wall_top[0];

		memmove(&game->wall_top[2]   , &game->wall_top[3]   , size);
		memmove(&game->wall_bottom[2], &game->wall_bottom[3], size);
	}

	/*
		Remove old obstacles
	*/
	if (game->obstacle_count > 0 && game->obstacle[0].x + game->obstacle[0].width < 0)
	{
		game->obstacle_count -= 1;
		memmove(&game->obstacle[0], &game->obstacle[1],
		        game->obstacle_count * sizeof game->obstacle[0]);
	}

	/*
		Update worm
	*/
	update_worm(game);

	/*
		Add forward wall point if required

		Changing probabilities here change how the cave
		looks
	*/
	if (game->wall_top[game->wall_count - 1].x < GAME_WIDTH)
	{
		XPoint * const top_point = &game->wall_top[game->wall_count];
		XPoint * const bottom_point = &game->wall_bottom[game->wall_count];

		game->wall_count += 1;

		top_point->x = game_random(5 + game->wall_height,
		                           40 + game->wall_height)
		                           + GAME_WIDTH;

		top_point->y = game_random(0, game->wall_height);

		bottom_point->x = top_point->x;
		bottom_point->y = top_point->y + game->tunel_height;
	}

	/*
		Add one new obstacles if required.
	*/
	if (game->obstacle_milestone == 0)
	{
		const unsigned int idx = game->obstacle_count;

		game->obstacle_milestone = game_random(20, 40);

		game->obstacle[idx].x = GAME_WIDTH;

		/* we should make it in the tunel, always */
		game->obstacle[idx].y = game_random(0, GAME_HEIGHT);

		game->obstacle[idx].width  = OBSTACLE_WIDTH;
		game->obstacle[idx].height = OBSTACLE_HEIGHT;

		game->obstacle_count += 1;
	}

	/*
		Make the decision if we are dead or not.

		- Check wall death
		- Check obstacle death
	*/

	/* wall death */
	{
		Region empty, final;
		Region tmp;
		Region up, down;
		XRectangle rect;

		/*
			We create a region with the two polygons.
			However, we also need to add a minimum
			off-screen rectangle region for top and
			bottom since the polygons are thinner then
			5 pixel height (required for death).
		*/

		empty = XCreateRegion();
		final = XCreateRegion();

		up   = XPolygonRegion(game->wall_top, game->wall_count, EvenOddRule);
		down = XPolygonRegion(game->wall_bottom, game->wall_count, EvenOddRule);
		XUnionRegion(up, down, final);
		XDestroyRegion(up);
		XDestroyRegion(down);

		tmp = XCreateRegion();
		XUnionRegion(empty, final, tmp);    /* copy final in tmp */

		rect.x = 0;
		rect.y = -5; 
		rect.width = GAME_WIDTH;
		rect.height = 5;
		XUnionRectWithRegion(&rect, tmp, final);

		XUnionRegion(empty, final, tmp);    /* copy final in tmp */

		rect.x = 0;
		rect.y = GAME_HEIGHT; 
		rect.width = GAME_WIDTH;
		rect.height = 5;
		XUnionRectWithRegion(&rect, tmp, final);

		XDestroyRegion(tmp);

		if (XRectInRegion(final, WORM_LENGTH, game->worm_y, 1, 4) == RectangleIn)
		{
			game->game_over = True;
		}

		XDestroyRegion(empty);
		XDestroyRegion(final);

		if (game->game_over == True)
		{
			return;
		}
	}

	/* obstacle death */
	for (i = 0; i < game->obstacle_count; ++i)
	{
		XRectangle * const rectangle = &game->obstacle[i];
		const unsigned int x = WORM_LENGTH, y = game->worm_y;

		/*
			We are testing if (x,y) is in the rectangle.

			If yes, it must have 4 pixel under it which
			are part of the rectangle rectangle to be
			dead.
		*/
		if (x > rectangle->x && x < rectangle->x + rectangle->width
		&&  y > rectangle->y && y < rectangle->y + rectangle->height
		&&  y+3 >= rectangle->y && y+3 < rectangle->y + rectangle->height)
		{
			game->game_over = True;
			return;    /* we are done with this game */
		}

		if (rectangle->x > WORM_LENGTH)
		{
			break;
		}
	}
}

static unsigned long
diff_timeval(const struct timeval *time1, const struct timeval *time0)
{
	unsigned long total;

	assert(time1->tv_sec >= time0->tv_sec);
	/* We are asserting that if a = b then c >= d. That is, a=b => c>d */
	/* both are equivalent :) */
	assert(time1->tv_sec == time0->tv_sec ? time1->tv_usec >= time0->tv_usec ? 1 : 0 : 1);
	assert((time1->tv_sec == time0->tv_sec && time1->tv_usec >= time0->tv_usec) || time1->tv_sec != time0->tv_sec);

	total = (time1->tv_sec - time0->tv_sec) * 1000000;

	if (total > 0)
	{
		total -= 1000000;
		total += 1000000 - time0->tv_usec;
		total += time1->tv_usec;
	}
	else
	{
		total += time1->tv_usec - time0->tv_usec;
	}

	return total;
}

/*
	Return a random number in the range [n,m)
*/
static unsigned int
game_random(unsigned int n, unsigned int m)
{
double foo = n + ((double) longrand() / 2147483646L * m);
	assert(n < m);

	/*printf("%f\n", foo);*/
	return foo;
}

/*
	At start the array isn't full. game->worm_count is slowly
	incremented. Once the array is full, the use of modulo make
	the oldest element being overwrited by the newest. The worm
	segment array is cycled that way.
*/
static void
update_worm(GameClass *game)
{
	int idx = game->tick_count;

	if (idx != 0)
	{
		idx %= WORM_LENGTH;
	}

	game->worm[idx].x1 = WORM_LENGTH - 1;
	game->worm[idx].x2 = WORM_LENGTH - 1;
	game->worm[idx].y1 = game->worm_y;
	game->worm[idx].y2 = game->worm_y + 4;

	if (game->worm_count < WORM_LENGTH)
	{
		game->worm_count += 1;
	}
}

/*
**  longrand() -- generate 2**31-2 random numbers
**
**  public domain by Ray Gardner
** 
**  based on "Random Number Generators: Good Ones Are Hard to Find",
**  S.K. Park and K.W. Miller, Communications of the ACM 31:10 (Oct 1988),
**  and "Two Fast Implementations of the 'Minimal Standard' Random
**  Number Generator", David G. Carta, Comm. ACM 33, 1 (Jan 1990), p. 87-88
**
**  linear congruential generator f(z) = 16807 z mod (2 ** 31 - 1)
**
**  uses L. Schrage's method to avoid overflow problems
*/
#define a 16807         /* multiplier */
#define m 2147483647L   /* 2**31 - 1 */

static long
nextlongrand(long seed)
{
      unsigned long lo, hi;

      lo = a * (long)(seed & 0xFFFF);
      hi = a * (long)((unsigned long)seed >> 16);
      lo += (hi & 0x7FFF) << 16;
      if (lo > m)
      {
            lo &= m;
            ++lo;
      }
      lo += hi >> 15;
      if (lo > m)
      {
            lo &= m;
            ++lo;
      }
      return (long)lo;
}

long randomnum = 1;

static long
longrand(void)                     /* return next random long */
{
      randomnum = nextlongrand(randomnum);
      return randomnum;
}

const struct GameClassPublicMethods Game =
{
	game_create,
	game_destroy,
	game_next_tick_in,
	game_score,
	game_score_replay,
	game_update_ticks,
};
