/* $Id: xsfcave.c,v 1.43 2006/01/31 01:16:17 ppl Exp $ */

/*
	Control the game. Enter the different states (intro/game/death).
*/

#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "xsfcave.h"

#include "field.h"
#include "game_engine.h"

#define WIDTH  (GAME_WIDTH  +  0)    /* top window width  */
#define HEIGHT (GAME_HEIGHT + 40)    /* top window height */

/*
	Information about the X server
*/
Display *display;
Window root;
int screen;
unsigned int depth;

int score;
int high_score;

enum { SPACE_UNKNOWN, SPACE_IS_PRESS, SPACE_IS_RELEASE } spacebar = { SPACE_UNKNOWN };

struct atoms atoms;

/*
	Command-line options.
*/
Bool fflag = False;
Bool kflag = False;
Bool lflag = False;
Bool mflag = False;
Bool nflag = False;
const char *nflag_arg;
Bool tflag = False;
Bool vflag = False;

#ifdef DEBUG
Bool wflag = False;
double wflag_arg;
#endif

static unsigned best_score_ever(void);
static void create_xsfcave_folder(void);
static void discard_button_press(Window);
static void draw_circle(Drawable, GC, unsigned int x, unsigned int y,
                        unsigned int diameter);
static void frame_save(int, Drawable, unsigned, unsigned);
static void score_draw(GameClass *, Drawable, GC fill_gc, GC text_gc);
static void signal_chld(int);
static void usage(char *, FILE *);
static const char *xsfcave_load_nickname(void);
static Bool xsfcave_save_nickname(const char *);
static Bool wait_event(XEvent *, unsigned int sec, unsigned int usec);
static void wait_user_response();

#ifdef DEBUG
	#define DEBUG_GETOPT "fw:z:"
#else
	#define DEBUG_GETOPT
#endif

int
main(int argc, char **argv)
{
	const char *display_name = NULL;
	const char *nickname;

	Window top_window;

	GameClass game;
	FieldClass field;

	GC red_gc;
	GC fill_gc;
	GC empty_gc;
	GC score_gc;

	Pixmap double_buffer;

	int i;
	int c;

	XEvent e;
	XGCValues values;

	XFontStruct *font;

	XSizeHints normal_hints;
	XWMHints wm_hints;

	ReplayClass replay;

	enum
	{
		IntroState,
		GameState,
		EndGameState,
		EndState
	} state;

	/*
		We have some command-line options here.
	*/
	while ( (c = getopt(argc, argv, "d:hklmn:tv" DEBUG_GETOPT)) != -1)
	{
		extern char *optarg;

		switch (c)
		{
		case 'd':
			display_name = optarg;
			break;

		case 'f':
			fflag = True;
			break;

		case 'h':
			usage(argv[0], stdout);
			exit(EXIT_SUCCESS);
			break;

		case 'k':
			kflag = True;
			break;

		case 'l':
			lflag = True;
			break;

		case 'm':
			mflag = True;
			break;

		case 'n':
			nflag = True;
			nflag_arg = optarg;
			break;

		case 't':
			tflag = True;
			break;

		case 'v':
			if (vflag == True)
			{
				int status;
				puts("Connecting to version server, please wait...");
				if (version_check() == -1)
				{
					puts("Failed to retrieve version information, try again later");
					exit(EXIT_FAILURE);
				}
				wait(&status);
				if (status != 0)
				{
					puts("Failed to retrieve version");
					exit(EXIT_FAILURE);
				}
				exit(EXIT_SUCCESS);
			}
			else
			{
				vflag = True;
			}
			break;

		#ifdef DEBUG
		case 'w':
			wflag = True;
			wflag_arg = atof(optarg);
			break;
		#endif

		case '?':
		default:
			usage(argv[0], stderr);
			exit(EXIT_FAILURE);
			break;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 0 && argc != 1)
	{
			usage(argv[0], stderr);
			exit(EXIT_FAILURE);
	}

	if (vflag == True)
	{
		puts(PACKAGE " " VERSION);
		exit(EXIT_SUCCESS);
	}

	create_xsfcave_folder();

	/*
		Controls the nickname use for submission. If
		"XSFCAVE_NICKNAME" has zero length, submission will
		not be triggered.
	*/
	if (nflag == True)
	{
		xsfcave_save_nickname(nflag_arg);
		nickname = nflag_arg;
	}
	else
	{
		nickname = xsfcave_load_nickname();
	}

	/*
		Open a connection with the X11 server.
	*/
	if ( (display = XOpenDisplay(display_name)) == NULL)
	{
		fprintf(stderr, "unable to open display \"%s\"\n", XDisplayName(NULL));
		exit(EXIT_FAILURE);
	}

	/*
		Register a signal handler to prevent zombie process
		accumulation.
	*/
	signal(SIGCHLD, signal_chld);

	/*
		Initialize global variables and seed
	*/
	root = DefaultRootWindow(display);

	screen = DefaultScreen(display);
	depth  = DefaultDepth(display, screen);

	if (atoms_init() == -1)
	{
		puts("Failed to intern required atoms");
		exit(EXIT_FAILURE);
	}

	/*
		Set a random seed for this session.
	*/
	randomnum = time(NULL) ^ getpid();

	#ifdef DEBUG
	XSynchronize(display, True);
	#endif

	/* Fill background color */
	values.fill_style = FillSolid;
	values.foreground = get_color(COLOR_PURPLE);
	fill_gc = XCreateGC(display, root, GCForeground|GCFillStyle, &values);

	values.foreground = get_color("red");
	red_gc = XCreateGC(display, root, GCForeground, &values);

	/*
		Create the score graphic context
	*/
	if ( (font = XLoadQueryFont(display, "-misc-fixed-*-*-*-*-14-*-*-*-*-*-*-*")) == NULL)
	{
		puts("Fixed font not available!");
		exit(EXIT_FAILURE);
	}
      
	values.font       = font->fid;
	values.foreground = WhitePixel(display, screen);

	score_gc = XCreateGC(display, root, GCForeground|GCFont, &values);

	empty_gc = XCreateGC(display, root, 0, &values);

	/*
		Create double buffering pixmap. Eventually we should
		check if the DBE extension is available on the X11
		server and use this instead.
	*/
	if ( (double_buffer = XCreatePixmap(display, root, WIDTH,
	                                    HEIGHT, depth)) == None)
	{
		puts("Failed to create double buffering pixmap!");
		XCloseDisplay(display);
		exit(EXIT_FAILURE);
	}

	/*
		Create all windows and set them
	*/
	top_window = XCreateSimpleWindow(display, root, 0, 0, WIDTH,
	                                 HEIGHT, 2,
	                                 WhitePixel(display, screen),
	                                 get_color(COLOR_PURPLE));

	/*
		Select the Xlib events we want to receive on our
		top-level window.
	*/
	XSelectInput(display, top_window, NoEventMask
	                                | ButtonPressMask
	                                | ExposureMask
	                                | KeyPressMask
	                                | KeyReleaseMask
	                                | PropertyChangeMask);

	normal_hints.flags      = PMinSize | PMaxSize;
	normal_hints.min_width  = normal_hints.max_width  = WIDTH;
	normal_hints.min_height = normal_hints.max_height = HEIGHT;
	XSetWMNormalHints(display, top_window, &normal_hints);

	wm_hints.flags = StateHint;
	wm_hints.initial_state = NormalState;

	if (kflag)
	{
		wm_hints.flags |= InputHint;
		wm_hints.input = False;
	}

	XSetWMHints(display, top_window, &wm_hints);

	if (mflag)
	{
		cursor_hide(top_window);
	}

	XStoreName(display, top_window, "xsfcave");

	XChangeProperty(display, top_window, atoms.WM_PROTOCOLS, atoms.ATOM, 32,
	                PropModeReplace, (void *) &atoms.WM_DELETE_WINDOW, 1);

	/*
		This is the decorationless window option.
	*/
	if (lflag == True)
	{
		CARD32 temp[5] = { 0x02 };

		XChangeProperty(display, top_window, atoms._MOTIF_WM_HINTS,
		                atoms._MOTIF_WM_HINTS, 32,
		                PropModeReplace, (void *) &temp, 5);
	}

	/*
		This is the window always on top option.
	*/
	if (tflag == True)
	{
		XChangeProperty(display, top_window, atoms._NET_WM_STATE,
		                atoms.ATOM, 32,
		                PropModeReplace, (void *) &atoms._NET_WM_STATE_ABOVE, 1);
	}

	XMapWindow(display, top_window);

	/*
		Replay from file and then exit()
	*/
	if (argc == 1)
	{
		FILE *fp;
		int i;
		int frame = 0;
		ReplayClass replay;

		Replay.create(&replay, randomnum);

		if ( (fp = fopen(argv[0], "r")) == NULL)
		{
			perror(argv[0]);
			exit(EXIT_FAILURE);
		}

		if (Replay.load(&replay, fp) == False)
		{
			fprintf(stderr, "Failed to load replay from \"%s\"\n", argv[0]);
			exit(EXIT_FAILURE);
		}

		fclose(fp);

		randomnum = replay.seed;

		if (Field.create(&field, top_window) == False)
		{
			puts("Failed to create game field!");
			XCloseDisplay(display);
			exit(EXIT_FAILURE);
		}

		Game.create(&game);

		XFillRectangle(display, double_buffer, fill_gc,
		               0, 0, WIDTH, HEIGHT);

		/*
			User is playing the game.
		*/
		for (i = 0; i < replay.tick_count; )
		{
			Field.draw(&field, &game, double_buffer);
			score_draw(&game, double_buffer, fill_gc, score_gc);
			XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
			          WIDTH, HEIGHT, 0, 0);

			if (wait_event(&e, 0, Game.next_tick_in(&game)) == True)
			{
				if (e.type == Expose)
				{
					/* continue will make us re-draw */
				}

				continue;
			}

			if (fflag)
			{
				frame_save(++frame, double_buffer, WIDTH, HEIGHT);
			}

			Game.update_ticks(&game, replay.user_updown[i++] == 'U' ? True : False, NULL);

			if (game.game_over == True) /* shouldn't happen */
			{
				break;
			}

		}

		/* must add a tick for game over since its replay */
		Game.update_ticks(&game, True, NULL);

		if (fflag)
		{
			frame_save(++frame, double_buffer, WIDTH, HEIGHT);
		}

		/*
			Game over animation
		*/
		for (i = 1; i < 18; ++i)
		{
			draw_circle(double_buffer, red_gc, WORM_LENGTH - 1, game.worm_y + 2, 4 + (i*4));

			XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
			          WIDTH, HEIGHT, 0, 0);

			if (fflag)
			{
				frame_save(++frame, double_buffer, WIDTH, HEIGHT);
			}

			while (wait_event(&e, 0, 50000) == True)
			{
				if (e.type == Expose)
				{
					XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
					          WIDTH, HEIGHT, 0, 0);
				}
			}
		}

		discard_button_press(top_window);

		while (1)
		{
			if (wait_event(&e, 5, 0) == True)
			{
				if (e.type == Expose)
				{
					XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
					          WIDTH, HEIGHT, 0, 0);
				}
				else if (e.type == ButtonPress || (e.type == KeyPress && XLookupKeysym(&e.xkey, 0) == XK_space))
				{
					break;
				}
			}
		}

		Game.destroy(&game);
		Field.destroy(&field);
		Replay.destroy(&replay);

		exit(EXIT_SUCCESS);
	}

	/*
		This is the game flow until user exit.

		1 - Intro animation
		2 - Play game
		3 - Game over animation
	*/
	for (; ; )
	{
		Bool go_up;
		state = IntroState;

		discard_button_press(top_window);

		/*
			Intro animation
		*/
		do
		{
			XFillRectangle(display, double_buffer, fill_gc,
			               0, 0, WIDTH, HEIGHT);
			intro_draw_frame(double_buffer);
			XCopyArea(display, double_buffer, top_window,
			          empty_gc, 0, 0, WIDTH, HEIGHT,
			          0, 0);

			while (wait_event(&e, 0, 10000) == True)
			{
				if (e.type == Expose)
				{
					break;    /* re-draw ourself */
				}
				else if (e.type == ButtonPress || (e.type == KeyPress && XLookupKeysym(&e.xkey, 0) == XK_space))
				{
					state = GameState;
					break;
				}
			}

		} while (state == IntroState);

		/*
			Play game.
		*/
		if (Field.create(&field, top_window) == False)
		{
			puts("Failed to create game field!");
			XCloseDisplay(display);
			exit(EXIT_FAILURE);
		}

		Game.create(&game);
		Replay.create(&replay, randomnum);

		XFillRectangle(display, double_buffer, fill_gc,
		               0, 0, WIDTH, HEIGHT);

		/*
			User is playing the game.
		*/
		do
		{
			int d; /* d like dummy */
			Window y; /* y like ymmud */

			unsigned int mask;

			Field.draw(&field, &game, double_buffer);
			score_draw(&game, double_buffer, fill_gc, score_gc);
			XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
			          WIDTH, HEIGHT, 0, 0);

			if (wait_event(&e, 0, Game.next_tick_in(&game)) == True)
			{
				if (e.type == Expose)
				{
					/* continue will make us re-draw */
				}

				continue;
			}

			go_up = False;

			/*
				We should probably track
				ButtonPress/Release events. I've
				tryed it but there was a few bugs.
				For now this is acceptable.
			*/
			if (XQueryPointer(display, top_window, &y, &y, &d, &d, &d, &d, &mask) != False)
			{
				if (mask & (Button1Mask|Button2Mask|Button3Mask|Button4Mask|Button5Mask) )
				{
					go_up = True;
				}
			}

			if (spacebar == SPACE_IS_PRESS)
			{
				go_up = True;
			}

			Game.update_ticks(&game, go_up, &replay);

			if (game.game_over == True)
			{
				state = EndGameState;
			}

		} while (state == GameState);

		score = Game.score(&game);

		/*
			Save replays in ~/.xsfcave

				- Last replay
				- Highest score in current session replay
				- Highest score ever replay
		*/

		Replay.save_last(&replay);

		if (score >= high_score)
		{
			Replay.save_best(&replay);
		}

		if (score >= best_score_ever())
		{
			Replay.save_best_ever(&replay);
		}

		/*
			Send the score over the Internet.
		*/
		if (strlen(nickname) > 0)
		{
			ScreenshotClass ss;

			Field.draw(&field, &game, double_buffer);
			score_draw(&game, double_buffer, fill_gc, score_gc);
			for (i = 1; i < 18/2; ++i)
			{
				draw_circle(double_buffer, red_gc, WORM_LENGTH - 1, game.worm_y + 2, 4 + (i*4));
			}

			Screenshot.create(&ss);
			Screenshot.capture(&ss, double_buffer, WIDTH, HEIGHT);
			submit_score(nickname, VERSION, &replay, &ss);
			Screenshot.destroy(&ss);
		}

		/*
			All must be displayed before going into the
			end game animation
		*/
		Field.draw(&field, &game, double_buffer);
		score_draw(&game, double_buffer, fill_gc, score_gc);

		/*
			Game over animation
		*/
		for (i = 1; i < 18; ++i)
		{
			draw_circle(double_buffer, red_gc, WORM_LENGTH - 1, game.worm_y + 2, 4 + (i*4));

			XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
			          WIDTH, HEIGHT, 0, 0);

			while (wait_event(&e, 0, 50000) == True)
			{
				if (e.type == Expose)
				{
					XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
					          WIDTH, HEIGHT, 0, 0);
				}
			}
		}

		/*
			Display game over and special message if
			we did a high score
		*/

		discard_button_press(top_window);

		/*
			Wait for a button press or timeout
		*/
		do
		{
			if (wait_event(&e, 5, 0) == True)
			{
				if (e.type == Expose)
				{
					XCopyArea(display, double_buffer, top_window, empty_gc, 0, 0,
					          WIDTH, HEIGHT, 0, 0);
				}
				else if (e.type == ButtonPress || (e.type == KeyPress && XLookupKeysym(&e.xkey, 0) == XK_space))
				{
					state = EndState;
				}
			}
			else
			{
				state = EndState;
			}
		} while (state == EndGameState);

		Replay.destroy(&replay);
		Game.destroy(&game);

		Field.destroy(&field);

		high_score = score > high_score ? score : high_score;
	}

	/* Not reached */

	return 0;
}

/*
	Attempt to create ~/.xsfcave
*/
static unsigned
best_score_ever(void)
{
	char *home;
	const char filename[] = "/.xsfcave/" BESTEVER_FILENAME;
	char *fullpath;
	FILE *fp;
	ReplayClass replay;
	unsigned int score;

	if ( (home = getenv("HOME")) == NULL)
	{
		return 0;
	}

	if ( (fullpath = malloc(strlen(home) + sizeof filename)) == NULL)
	{
		perror("malloc");
		return 0;
	}

	sprintf(fullpath, "%s%s", home, filename);

	/*
		If this is the first time we play
		it will output an error
	*/
	if ( (fp = fopen(fullpath, "r")) == NULL)
	{
		perror(fullpath);
	}

	free(fullpath);

	if (fp == NULL)
	{
		return 0;
	}

	Replay.create(&replay, 0);

	if (Replay.load(&replay, fp) == False)
	{
		score = 0;
	}
	else
	{
		score = Game.score_replay(&replay);
	}

	fclose(fp);
	Replay.destroy(&replay);

	return score;
}

/*
	Attempt to create ~/.xsfcave
*/
static void
create_xsfcave_folder(void)
{
	char *home;
	const char filename[] = "/.xsfcave";
	char *fullpath;

	if ( (home = getenv("HOME")) == NULL)
	{
		return;
	}

	if ( (fullpath = malloc(strlen(home) + sizeof filename)) == NULL)
	{
		perror("malloc");
		return;
	}

	sprintf(fullpath, "%s%s", home, filename);

	if (mkdir(fullpath, 0755) == -1)
	{
		if (errno != EEXIST)
		{
			perror(fullpath);
		}
	}

	free(fullpath);

}

/*
	Draw a circle of a certain diameter with center (x,y).
*/
static void
draw_circle(Drawable d, GC gc, unsigned int x, unsigned int y, unsigned int diameter)
{
	XDrawArc(display, d, gc, x - diameter/2, y - diameter/2,
	         diameter, diameter, 360*64, 360*64);
}


/*
	Allocate color name or return black pixel as failure
*/
unsigned long
get_color(const char *color_name)
{
	XColor color;
	Colormap colormap;
    
	colormap = XDefaultColormap(display, screen);
 
	/*
		We only free on XCloseDisplay()
	*/
	if (XAllocNamedColor(display, colormap, color_name, &color,
	                     &color) == 0)
	{
		return BlackPixel(display, screen);
	}

	return color.pixel;
}

/*
	Wait for an event or timeout. We manage some events general
	to the entire application here.

	If we receive a lot of events, we might sleep a little more
	then timeout. This shouldn't matter in practice though.

	Pitfall: Easy to get flood of events. i.e. NoExpose

	Return True if `e' was set. Otherwise the timeout have
	passed.
*/
static Bool
wait_event(XEvent *e, unsigned int sec, unsigned int usec)
{
	const int xfd = ConnectionNumber(display);

	for (; ; )
	{
		int n;

		fd_set readfds;
		struct timeval val;

		while (XPending(display) > 0)
		{
			XNextEvent(display, e);

			switch (e->type)
			{
			case ClientMessage:
				/*
					One possible client message
					is when the user attemps
					to close the window. An
					example of this is when a
					user click the [X] of the
					window manager title bar.
				*/
				if (e->xclient.message_type == atoms.WM_PROTOCOLS
				&&  e->xclient.format == 32
				&&  e->xclient.data.l[0] == atoms.WM_DELETE_WINDOW)
				{
					XCloseDisplay(display);
					exit(EXIT_SUCCESS);
				}
				#ifdef DEBUG
				else
				{
					puts("unknown property notify");
				}
				#endif
				break;

			case Expose:
				while (XCheckWindowEvent(display,
				                         e->xexpose.window,
				                         ExposureMask, e)
				        == True)
				{
					/* discard out-dated events */
				}

				return True;
				break;

			case KeyPress:
				switch (XLookupKeysym(&e->xkey, 0))
				{
				case XK_q:
				case XK_Q:
				case XK_Escape:
					XCloseDisplay(display);
					exit(EXIT_SUCCESS);
					break;

				case XK_space:
					spacebar = SPACE_IS_PRESS;
					break;

				default:
					break;
				}
				return True;
				break;

			case KeyRelease:
				switch (XLookupKeysym(&e->xkey, 0))
				{
				case XK_space:
					spacebar = SPACE_IS_RELEASE;
					break;

				default:
					break;
				}
				return True;
				break;

			case NoExpose:
				/* we ignore this event (else we get overwhelmed) */
				break;

			default:
				return True;
				break;
			}
		}

		FD_SET(xfd, &readfds);

		val.tv_sec = sec;
		val.tv_usec = usec;
		n = select(xfd+1, &readfds, NULL, NULL, &val);

		/*
			The time request has pass and no events was
			received meanwhile.
		*/
		if (n == 0)
		{
			return False;
		}
	}
}

static void
discard_button_press(Window w)
{
	XEvent e;
	while (XCheckWindowEvent(display, w, ButtonPressMask, &e) == True)
	{
	}
}

static void
frame_save(int frame, Drawable d, unsigned width, unsigned height)
{
	FILE *fp;
	char filename[] = "frame-XXXXXX.bmp";
	ScreenshotClass ss;

	sprintf(filename, "frame-%06d.bmp", frame);
	if ( (fp = fopen(filename, "wb")) == NULL)
	{
		perror(filename);
		exit(EXIT_FAILURE);
	}

	Screenshot.create(&ss);
	Screenshot.capture(&ss, d, WIDTH, HEIGHT);
	Screenshot.save(&ss, fp);
	Screenshot.destroy(&ss);

	fclose(fp);
}

static void
score_draw(GameClass *game, Drawable d, GC fill_gc, GC text_gc)
{
	char buf[1024];

	XFillRectangle(display, d, fill_gc, 0, GAME_HEIGHT, WIDTH, HEIGHT - GAME_HEIGHT);

	sprintf(buf, "Score:%5d", (int) Game.score(game));
	XDrawString(display, d, text_gc, 15, GAME_HEIGHT + 20, buf, strlen(buf));
}

/*
	SIGCHLD signal handler.
	Discard all zombies.
*/
static void
signal_chld(int signo)
{
    int stat;

    while (waitpid(-1, &stat, WNOHANG) > 0)
    { 
    } 
      
    return;
}

/*
	Display the program usage message to the user.
*/
static void
usage(char *prog_name, FILE *fp)
{
	fprintf(fp, "usage: %s [-hkmtv] [-d <display>] [-n <nickname>] [-vv] [-w <weight>] [replay.xsfcave]\n", basename(prog_name));
	fputs("\t-d\tX11 display to connect to\n", fp);
	fputs("\t-h\tDisplay this help message and exit\n", fp);
	fputs("\t-k\tTurn off keyboard input\n", fp);
	fputs("\t-l\tUse a decorationless window\n", fp);
	fputs("\t-m\tHide mouse pointer\n", fp);
	fputs("\t-n\tEnable Internet score submission using a nickname\n", fp);
	fputs("\t-t\tRun always on top\n", fp);
	fputs("\t-v\tDisplay version and exit\n", fp);
	fputs("\t-vv\tDisplay current and latest version from the web and exit\n", fp);
}

static void
wait_user_response()
{
	(void)wait_user_response;
}

const char *
xsfcave_load_nickname(void)
{
	char *home;
	const char filename[] = "/.xsfcave/" NICKNAME_FILENAME;
	char *fullpath;
	FILE *fp;
	static char nickname[80] = "";
	char *rtn;

	if ( (home = getenv("HOME")) == NULL)
	{
		return nickname;
	}

	if ( (fullpath = malloc(strlen(home) + sizeof filename)) == NULL)
	{
		perror("malloc");
		return nickname;
	}

	sprintf(fullpath, "%s%s", home, filename);

	if ( (fp = fopen(fullpath, "r")) == NULL)
	{
		if (errno == ENOENT)
		{
			puts("You can register your scores in the Hall of Fame using the -n option");
		}
		else
		{
			perror(fullpath);
		}
	}

	free(fullpath);

	if (fp == NULL)
	{
		return nickname;
	}

	if ( (rtn = fgets(&nickname[0], sizeof nickname, fp)) == NULL)
	{
		perror("fgets");
	}

	fclose(fp);

	return nickname;
}

Bool
xsfcave_save_nickname(const char *nickname)
{
	char *home;
	const char filename[] = "/.xsfcave/" NICKNAME_FILENAME;
	char *fullpath;
	FILE *fp;
	Bool rtn = True;

	if ( (home = getenv("HOME")) == NULL)
	{
		return False;
	}

	if ( (fullpath = malloc(strlen(home) + sizeof filename)) == NULL)
	{
		perror("malloc");
		return False;
	}

	sprintf(fullpath, "%s%s", home, filename);

	if ( (fp = fopen(fullpath, "w")) == NULL)
	{
		perror(fullpath);
	}

	free(fullpath);

	if (fp == NULL)
	{
		return False;
	}

	if (fputs(nickname, fp) == EOF)
	{
		perror("fputs");
		rtn = False;
	}

	fclose(fp);

	return rtn;
}
